/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 * 
 * $Revision: 1.69 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org. Derived from vplay, written by
 * tridge@samba.org, December 2000
 * 
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifndef TIVO
#include <netinet/tcp.h>
#endif
#include "playitsam.h"
#include "proto.h"
#ifdef TIVO
#include "libmfs/mfs.h"
#include "osd.h"
#include "tproto.h"
int snprintf(char *str, size_t size, const char *format,...);
#endif
#ifdef MPEG2DEC
#include "mpeg2dec/video_out.h"
#include "mpeg2dec/mpeg2.h"
#include "mpeg2dec/mm_accel.h"
#endif

#ifdef TIVO
/* For multi-fsid recordings, we store
 * the fsids and their sizes here.
 */
struct fsid_group {
  int fsid;
  int start_chunk;
  int end_chunk;
} *Fsid_list=NULL;
int Fsid_list_count=0;
#endif

 

FILE *audio_f;			/* Non-TiVo audio pipe */
FILE *video_f;			/* Non-TiVo video pipe */
int audio_fd;			/* fd for audio device */
int video_fd;			/* fd for video device */
int stream_fd;			/* fd where tystream comes in */
char *outfname= NULL;		/* Output (recording) file name */

#ifdef MPEG2DEC
mpeg2dec_t mpeg2dec;
vo_instance_t vo_out;
static int disable_accel= 0;
#endif

int verbose= 0;			/* Does user want verbose output */
int allow_playback= 1;		/* Are we going to play to devices? */
				/* If ==2, then only play, no editing */
int seeking_ok= 1;		/* Can we seek from stream file? */
int chunk, totalchunks;		/* Current and total chunks */

int exit_things= 0;		/* Things to do at exit time */
struct termios oldtios, newtios;/* Console settings for raw & cooked modes */
				/* Buffers for two chunks */
u_int8_t inbuf1[CHUNK_SIZE], inbuf2[CHUNK_SIZE];
u_int8_t *buf=inbuf1;

int inputtype= 0;		/* Type of file input */

int do_osd= 1;			/* Do we want on-screen display? */

int osd_x= 1, osd_y= 15;	/* Co-ords for OSD string display */
int osd_dirty= 0;		/* Is there any text on the screen? */
int osd_chunk= 0;		/* If >0, this chunk # now osd */
char osd_string[80];		/* String to draw on the screen */

#ifdef TIVO
/* Read from a single fsid or from a set of fsids. Always add 1
 * so as to skip the chunk #0 in every fsid stream.
 */
int mfs_multifsid_pread(int fsid, char *buf, int chunknum)
{
 int i;

  /* If we only have one fsid, use the one supplied */
  if (Fsid_list_count==0) {
    return( mfs_fsid_pread(fsid, buf, ((u64) chunknum+1) << 17, CHUNK_SIZE) );
  }
  

  /* Otherwise, search for an fsid that matches and use it */
  for (i=0; i<Fsid_list_count; i++) {
    if ((chunknum >=Fsid_list[i].start_chunk) &&
        (chunknum <=Fsid_list[i].end_chunk)) {
	/* fprintf(stderr,"Reading from fsid %d\r\n",Fsid_list[i].fsid); */
          return( mfs_fsid_pread(Fsid_list[i].fsid, buf,
		 ((u64) (1+ chunknum - Fsid_list[i].start_chunk)) << 17,
								CHUNK_SIZE) );
    }
  }
  return(-1);	/* No match */
}
#endif



/*
 * On exit, close the various input and output devices, and reset the
 * keyboard to its usual mode.
 */
void 
Exit(int how)
{
  if (exit_things & EXIT_CLOSEAVDEVS) {
#ifdef MPEG2DEC
    mpeg2_close(&mpeg2dec);
#else
    close(video_fd);
#endif
    close(audio_fd);
    if (audio_f) fclose(audio_f);
    if (video_f) fclose(video_f);
  }
  if (exit_things & EXIT_FIXKEYBOARD) tcsetattr(0, TCSANOW, &oldtios);

#ifdef TIVO
  if (exit_things & EXIT_CLOSEREMDEV)
    close_remote();
  if (do_osd) { ClearOSD(); DrawOSD(); FreeTextOSD(); }
  if ((how == 0) && (exit_things & EXIT_MYWORLD)) start_myworld();
#endif

  printf("\n\n");
  close(0);
  exit(how);
}



/* open a socket to a tcp remote host with the specified port */
int
open_socket_to_tyserver(char *inhost, int port)
{
  int type= SOCK_STREAM;
  struct sockaddr_in sock_out;
  int res;
  int nodelay= 1;
  int window= 65535;
  char *host, *numptr, *numend;
  u_int32_t raw_address= 0;
  int one_part;

  host= strdup(inhost);
  if (host == NULL) return (-1);

  res= socket(PF_INET, type, 0);
  if (res == -1) {
    free(host);
    return (-1);
  }
#ifdef CAN_DO_NAME_LOOKUPS
  /* Can someone tell me how to get /etc/hosts working on the TiVo? */
  hp= gethostbyname(host);
  if (!hp) {
    fprintf(stderr, "unknown host: %s\n", host);
    return (-1);
  }
  memcpy(&sock_out.sin_addr, hp->h_addr, hp->h_length);
#else

  /* Walk the IP address and convert into a 32-bit integer */
  numptr= host;
  numend= strchr(host, '.');
  if (numend == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend= '\0';
  one_part= atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address= one_part << 24;

  numptr= ++numend;
  numend= strchr(numptr, '.');
  if (numptr == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend= '\0';
  one_part= atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += (one_part << 16);

  numptr= ++numend;
  numend= strchr(numptr, '.');
  if (numptr == NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  *numend= '\0';
  one_part= atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += (one_part << 8);
  numptr= ++numend;
  one_part= atoi(numptr);
  if ((one_part < 0) || (one_part > 255)) {
    fprintf(stderr, "%s not an IP address\n", inhost);
    free(host); return (-1);
  }
  raw_address += one_part;
  sock_out.sin_addr.s_addr= htonl(raw_address);
#endif

  sock_out.sin_port= htons(port);
  sock_out.sin_family= PF_INET;

  if (connect(res, (struct sockaddr *)&sock_out, sizeof(sock_out))) {
    close(res);
    fprintf(stderr, "failed to connect to %s\n", inhost);
    free(host); return (-1);
  }
  setsockopt(res, SOL_SOCKET, SO_RCVBUF, &window, sizeof(int));
  setsockopt(res, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));

  free(host); return (res);
}


/* Set a fd into nonblocking mode */
void set_nonblocking(int fd)
{
  int val;

  if ((val= fcntl(fd, F_GETFL, 0)) == -1) return;
  if (!(val & O_NONBLOCK)) { val |= O_NONBLOCK; fcntl(fd, F_SETFL, val); }
}


/*
 * On a non-TiVo, we need to start up external processes to play a stream
 * back. We use popen and then get the file descriptors for the two pipes.
 */
void
openpipes(char *audio_prog, char *video_prog)
{
#ifdef MPEG2DEC
  u_int32_t accel;
  vo_driver_t *vo_driver;
  vo_driver= vo_drivers();
#endif

  audio_f= popen(audio_prog, "w");
  if (!audio_f) {
    fprintf(stderr, "Failed to start audio program %s\n", audio_prog);
    Exit(1);
  }
#ifdef MPEG2DEC
  accel= disable_accel ? 0 : (mm_accel() | MM_ACCEL_MLIB);
  vo_accel(accel);
  mpeg2_init(&mpeg2dec, accel, vo_open(vo_driver[0].open));
  video_fd= -1;			/* Set this otherwise chunk.c thinks */
				/* that we are recording, not playing */
#else
  video_f= popen(video_prog, "w");
  if (!video_f) {
    fprintf(stderr, "Failed to start video program %s\n", video_prog);
    Exit(1);
  }
  video_fd= fileno(video_f);
#endif
  audio_fd= fileno(audio_f);
  exit_things |= EXIT_CLOSEAVDEVS;
}


/*
 * Read a specific chunk from the input stream. Return 1 if ok, 0 if failure
 */
int
read_a_chunk(int fd, int chunknum)
{
  char *bptr= buf;
  int numread= 0;
  int numleft= CHUNK_SIZE;

  switch (inputtype) {
#ifdef TIVO
  case INPUT_FSID:
    numread= mfs_multifsid_pread(fd, buf, chunknum);
    if (numread != CHUNK_SIZE) {
      fprintf(stderr, "Failure reading chunk %d\r\n", chunknum);
      Exit(1);
    }
    return (1);
#endif
  case INPUT_STDIN:
  case INPUT_FILE:
  case INPUT_TYSERVER:
    /*
     * We only seek if we can. This means that the forward/backward commands
     * are NOT going to work for input on stdin.
     */
    if (seeking_ok) lseek(fd, CHUNK_SIZE * chunknum, SEEK_SET);

    /* Loop until we do get CHUNK_SIZE bytes of data */
    while (numleft > 0) {
      numread= read(fd, bptr, numleft);
      if (numread < 1) {
	fprintf(stderr, "Failure Reading chunk %d\r\n", chunknum);
	Exit(1);
      }
      numleft -= numread; bptr += numread;
    }
    return (1);
  default:
    return (0);
  }
}



/*
 * Obtain a command from the user and return it. Commands can come in from
 * stdin or the TiVo remote.
 */
int
read_a_command(void)
{
  char ch;

#ifdef TIVO
  int key;

  /*
   * See if there is a keystroke from the TiVo remote control before we try
   * the telnet keyboard.
   */
  key= read_a_remote_command();

  switch (key) {
    case REM_TIVO_CENTRAL: return (CMD_QUIT);
    case REM_RECORD: return (CMD_RECORD);
    case REM_SELECT: return (CMD_PLAYITALL);
    case REM_PLAY: return (CMD_NOPAUSE);
    case REM_BACK: return (CMD_BACKFRAME);
    case REM_PAUSE: return (CMD_PAUSE);
    case REM_FORWARD: return (CMD_FWDFRAME);
    case REM_BACK30: return (CMD_START);
    case REM_END: return (CMD_END);
    case REM_1: return (CMD_FASTF1);
    case REM_2: return (CMD_FASTF2);
    case REM_3: return (CMD_FASTF3);
    case REM_4: return (CMD_FASTF4);
    case REM_5: return (CMD_FASTF5);
    case REM_6: return (CMD_FASTF6);
    case REM_7: return (CMD_FASTB2);
    case REM_8: return (CMD_FASTB3);
    case REM_9: return (CMD_FASTB4);
    case REM_CLEAR: return (CMD_CSTART);
    case REM_ENTER: return (CMD_CEND);
    case REM_THUMBS_UP: return (CMD_TOGSEECUT);
    default: break;
  }
#endif

  if (allow_playback == 2) return (CMD_NONE);	/* No keyboard cmds with -p */
  if (inputtype == INPUT_STDIN) return (CMD_NONE); /* No cmds, only stream */

  if (read(0, &ch, 1) != 1) return (CMD_NONE);	/* No kybrd data at present */

  switch (ch) {
    case ' ': return (CMD_PAUSE);
    case 'f': return (CMD_FWDFRAME);
    case 'b': return (CMD_BACKFRAME);
    case 3:			/* ctrl-C and q */
    case 'Q':
    case 'q': return (CMD_QUIT);
    case 's':
    case 'c': return (CMD_CSTART);
    case 'e': return (CMD_CEND);
    case 'R': return (CMD_RECORD);
    case 'P': return (CMD_PLAYITALL);
    case '1': return (CMD_FASTF1);
    case '2': return (CMD_FASTF2);
    case '3': return (CMD_FASTF3);
    case '4': return (CMD_FASTF4);
    case '5': return (CMD_FASTF5);
    case '6': return (CMD_FASTF6);
    case '!': return (CMD_FASTB1);
    case '@': return (CMD_FASTB2);
    case '#': return (CMD_FASTB3);
    case '$': return (CMD_FASTB4);
    case '%': return (CMD_FASTB5);
    case '^': return (CMD_FASTB6);
    case '0': return (CMD_END);
    case ')': return (CMD_START);
    case 'k': return (CMD_KILLCUT);
    case 'K': return (CMD_KILLALL);
    case 'l': return (CMD_LISTCUTS);
    case 'v': return (CMD_TOGSEECUT);
  }
  return (CMD_NONE);
}



/*
 * Given the input argument from the user, work out which one it is and open
 * it. Returns a file descriptor to the input, or a fsid if the input is on
 * MFS.
 */
int
open_input(char *inputsource)
{
#ifdef TIVO
  int fsid;
  int i;
  char *onefsid;
#endif
  int fd;
  struct stat statbuf;
  char ch, *tyfile;
  char result[4];

  /* Use stdin if the input source is "-" */
  if (!strcmp(inputsource, "-")) {
    seeking_ok= 0; totalchunks= CHUNKSUNKNOWN; inputtype= INPUT_STDIN;
    return (0);
  }
  /* If the input source has a colon, get it from the tyserver */

  if ((tyfile= strchr(inputsource, ':')) != NULL) {
    *tyfile= '\0'; tyfile++;
    fd= open_socket_to_tyserver(inputsource, 1150);
    if (fd == -1) Exit(1);

    /* Now send a PLAY command, and get a 200 reply */
    write(fd, "PLAY ", 5); write(fd, tyfile, strlen(tyfile));
    write(fd, "\r\n", 2); read(fd, result, 3);
    result[3]= '\0';
    if (strcmp(result, "200")) {		/* Did not get ok message */
      close(fd); fprintf(stderr, "Couldn't play from tyserver\n"); Exit(1);
    }
    /* Otherwise, read up to \n from tyserver before we get the stream */
    while (1) {
      read(fd, &ch, 1);
      if (ch == '\n') break;
    }
    seeking_ok= 0; totalchunks= CHUNKSUNKNOWN;
    inputtype= INPUT_TYSERVER; return (fd);
  }

#ifdef TIVO
# define MAX_FSID_LIST 50

   /**** BIG CHANGE FROM PREVOUS VERSIONS ****/
   /* We now always ignore chunk 0 from every fsid stream */

  /* inputsource could be a single fsid or a list of fsids */
  /* First, check if it has commas in it */
  /* If so, split the list into fsids, store them */
  /* and work out their size */
  if (strchr(inputsource,',')) {

    /* Malloc room for MAX_FSID_LIST entries */
    Fsid_list= (struct fsid_group *)malloc(
				MAX_FSID_LIST * sizeof(struct fsid_group));
    if (Fsid_list==NULL) { fprintf(stderr,"Can't malloc fsid list\n"); exit(1);}

    mfs_init(); totalchunks=0; fsid=0;
    /* Extract each fsid */
    for (i=0;(i<MAX_FSID_LIST)&&((onefsid=strsep(&inputsource,","))!=NULL);i++){
      fsid= atoi(onefsid);
      if (fsid<=0) continue;
      Fsid_list[i].fsid=fsid;
      Fsid_list[i].start_chunk=totalchunks;
      totalchunks+= mfs_fsid_numchunks(fsid) -1; /* Don't count chunk #0 */
      Fsid_list[i].end_chunk=totalchunks-1;
      Fsid_list_count++;
      fprintf(stderr, "fsid %d has %d chunks [%d to %d]\r\n", fsid, totalchunks,
			Fsid_list[i].start_chunk, Fsid_list[i].end_chunk);
    }
    seeking_ok= 0; inputtype= INPUT_FSID; return(fsid);
  }

  /* Nope, no commas, but if the input source is numeric, treat it as a fsid */
  fsid= atoi(inputsource);
  if (fsid != 0) {
    mfs_init();
    totalchunks= mfs_fsid_numchunks(fsid) -1;	/* Skip chunk 0 */
    fprintf(stderr, "fsid %d has %d chunks\r\n", fsid, totalchunks);
    seeking_ok= 0; inputtype= INPUT_FSID; return(fsid);
  }
#endif

  /* None of the above, treat it as a Unix filename */
  fd= open(inputsource, O_RDONLY);
  if (fd == -1) {
    fprintf(stderr, "Cannot open input file %s\n", inputsource); Exit(1);
  }
  if (stat(inputsource, &statbuf) == -1) {
    fprintf(stderr, "Cannot determine size of %s\n", inputsource); Exit(1);
  }
  totalchunks= statbuf.st_size / CHUNK_SIZE;
  seeking_ok= 1; inputtype= INPUT_FILE; return (fd);
}

#ifdef TIVO
void OsdString(char *str)
{
 osd_x= 1, osd_y= 15; ClearOSD();
 DrawString(&osd_x, &osd_y, str, 4, 1); DrawOSD();
 osd_dirty=1;
}
#endif


/*
 * Playback a program to the audio/video devices, removing the cut chunks,
 * and preventing editing from occurring. Also give delay before starting. fd
 * is stream input.
 */
void
playback_program(int fd)
{
  int chk, i;
  int cmd, give_up_level= 0;

  fprintf(stderr, "\nStarting to playback in\n");
#ifdef TIVO
  if (do_osd) { OsdString("Playback starting ..."); sleep(1);
  }
#endif

  for (i= 10; i > 0; i--) {
    fprintf(stderr, "%d seconds ...\n", i);
#ifdef TIVO
    if (do_osd) {
	snprintf(osd_string, 79, "%d seconds  ", i);
	OsdString(osd_string);
    }
#endif
    sleep(1);
  }

#ifdef TIVO
  if (do_osd) { ClearOSD(); DrawOSD(); osd_dirty=0; }
#endif

  for (chk= 0; chk < totalchunks; chk++) {
    read_a_chunk(fd, chk);
    edit_and_play_chunk(buf, chk, audio_fd, video_fd, 0, 1);

    /*
     * Give the user a chance to interrupt the playback by sending TiVo
     * Central, Select, TiVo Central with the remote
     */
    cmd= read_a_command();
    switch (cmd) {
      case CMD_QUIT: if (give_up_level == 0) give_up_level= 1;
      		     if (give_up_level == 2) Exit(0); break;
      case CMD_PLAYITALL: if (give_up_level == 1) give_up_level= 2;
    }
  }
  fprintf(stderr, "Playback done...\n");
}

/*
 * Export a program to some output file, removing the cut chunks, and
 * preventing editing from occurring. fd is stream input.
 */
void
record_program(char *outfile, int fd)
{
  int chk;
  int outfd;
  char *tyfile;
  char ch, result[4];

  if (outfile == NULL) { fprintf(stderr, "No output file was named\n");return; }
  				/* - means standard output */
  if (!strcmp(outfile, "-")) { outfd= 1; }
  		/* a name with a colon means to send it to the tyserver */
  else if ((tyfile= strchr(outfile, ':')) != NULL) {
    *tyfile= '\0';
    tyfile++;
    outfd= open_socket_to_tyserver(outfile, 1150);
    if (outfd == -1) return;

    		/* Now send a SAVE command, and get a 200 reply */
    write(outfd, "SAVE ", 5); write(outfd, tyfile, strlen(tyfile));
    write(outfd, "\r\n", 2); read(outfd, result, 3);
    result[3]= '\0';
    if (strcmp(result, "200")) {		/* Did not get ok message */
      close(outfd); fprintf(stderr, "Couldn't record to tyserver\n"); return; }
    		/* Otherwise, read up to \n from tyserver before we send */
    while (1) {
      read(outfd, &ch, 1);
      if (ch == '\n') break;
    }
  }
  /* Otherwise, assume a Unix filename */
  else {
    outfd= open(outfile, O_WRONLY | O_CREAT | O_TRUNC, 0644);
    if (outfd== -1) { fprintf(stderr, "Cannot write to %s\n", outfile);return; }
  }

  fprintf(stderr, "\nStarting to record ...\n");
#ifdef TIVO
  if (do_osd) OsdString("Recording ...");
#endif

  for (chk= 0; chk < totalchunks; chk++) {

    /* Skip chunks that are completely outside a cut */
    if (find_chunk_in_cutlist(chk)==IS_INSIDE_CUT) continue;

    read_a_chunk(fd, chk);
    edit_and_play_chunk(buf, chk, outfd, 0, 0, 1);
    if (chk && ((chk % 100) == 0)) {
      fprintf(stderr, "Chunk %d\n", chk);
#ifdef TIVO
      if (do_osd) {
	snprintf(osd_string, 79, "Chunk %d", chk);
	OsdString(osd_string);
      }
#endif
    }
  }
  fprintf(stderr, "Recording done\n");
#ifdef TIVO
  if (do_osd) { ClearOSD(); DrawOSD(); osd_dirty=0; }
#endif
}

static void
edit_program(void)
{
  int cmd, cstart, is_a_cut= 0;
  int gop_indicator= 0;
  int paused= 0;		/* Start in non-paused mode */
  int skip_cuts=0; 		/* Toggle: if 1 we skip the cut chunks. */
  				/* If 0, we play all chunks */
  chunk= 0; buf=inbuf1;
  /*
   * Loop reading chunks from stream input, and dealing with user commands.
   * Play back the stream if allowed.
   */
  while (1) {
    /* Make sure that the chunk falls into acceptable limits */
    if (chunk < 1) chunk= 1;
    if (chunk >= totalchunks) { chunk= totalchunks - 1; paused= 1; }

    if (chunk < totalchunks) {
      read_a_chunk(stream_fd, chunk);
      gop_indicator= edit_and_play_chunk(buf, chunk,
			audio_fd, video_fd, 1, skip_cuts);
      /* Always swap buffers around */
      if (buf==inbuf1) buf=inbuf2; else buf=inbuf1;
    }

    /*
     * To stop editing from happening on a chunk without a GOP, we
     * immediately go back and play another frame. Not true if this is the
     * last chunk, however.
     */
    if ((gop_indicator!=0) && (chunk != (totalchunks-1))) { chunk++; continue; }
here:
    if (paused) {
      /*
       * Given that we are paused, let's find out if this chunk is a cut
       * start or end point, and display that.
       */
      is_a_cut= find_chunk_in_cutlist(chunk);
      switch (is_a_cut) {
        case IS_INSIDE_CUT:
        case IS_OUTSIDE_CUT: snprintf(osd_string, 79, "Chunk %d", chunk);
			     break;
        case IS_CUT_START: snprintf(osd_string, 79, "Cut start at chunk %d",
		 						chunk);
				break;
        case IS_CUT_END: snprintf(osd_string, 79, "Cut end at chunk %d",
		 						chunk);
				break;
      }
#ifdef TIVO
      if (do_osd && (osd_chunk != chunk)) {
        OsdString(osd_string); osd_chunk= chunk;
      }
#endif
      fprintf(stderr, "\r%s", osd_string);
    } else fprintf(stderr, "\rChunk %d", chunk);

    if (totalchunks != CHUNKSUNKNOWN) fprintf(stderr," out of %d", totalchunks);
    fprintf(stderr, "                     ");

    cmd= read_a_command();
    switch (cmd) {
      case CMD_NONE: break;
      case CMD_PAUSE: paused= 1 - paused; break;
      case CMD_NOPAUSE: paused= 0; break;
      case CMD_END: chunk= totalchunks - 1; paused= 1; break;
      case CMD_START: chunk= 1; break;
      case CMD_FWDFRAME: chunk++; break;
      case CMD_BACKFRAME: chunk -= 2; break;
      case CMD_FASTF1: chunk += 5; break;
      case CMD_FASTF2: chunk += 25; break;
      case CMD_FASTF3: chunk += 125; break;
      case CMD_FASTF4: chunk += 500; break;
      case CMD_FASTF5: chunk += 2500; break;
      case CMD_FASTF6: chunk += 12500; break;
      case CMD_FASTB1: chunk -= 5; break;
      case CMD_FASTB2: chunk -= 25; break;
      case CMD_FASTB3: chunk -= 125; break;
      case CMD_FASTB4: chunk -= 500; break;
      case CMD_FASTB5: chunk -= 2500; break;
      case CMD_FASTB6: chunk -= 12500; break;
      case CMD_TOGSEECUT: skip_cuts= 1-skip_cuts;
			  if (skip_cuts) snprintf(osd_string, 79,
					"Cut material will not be displayed");
			  else snprintf(osd_string, 79,
			    		"Cut material WILL not be displayed");
			  fprintf(stderr, "\r\n%s\r\n", osd_string);
#ifdef TIVO
			  if (do_osd) OsdString(osd_string);
#endif
			  break;

      /* On quit, reset the keyboard to cooked */
      case CMD_QUIT: Exit(0);

      case CMD_CSTART:
        if (gop_indicator!=0) break;
        add_cut_start(chunk);
        fprintf(stderr, "\r\nCut starting at chunk %d\r\n", chunk);
#ifdef TIVO
        if (do_osd && paused) {
	  snprintf(osd_string, 79, "Cut starting at %d", chunk);
	  OsdString(osd_string);
        }
#endif
        break;

      case CMD_CEND:
        if (gop_indicator!=0){fprintf(stderr, "\r\nCan't cut here, no GOP\r\n");
			  break; }
        if ((cstart= add_cut_end(chunk)) == -1) {
	  fprintf(stderr, "\r\nError ending chunk at this point\r\n");
	  break;
        }
        fprintf(stderr, "\r\nNew cut from %d to %d\r\n", cstart, chunk);
#ifdef TIVO
        if (do_osd && paused) {
	  snprintf(osd_string, 79, "New cut %d to %d", cstart, chunk);
	  OsdString(osd_string);
       }
#endif
        break;

      case CMD_RECORD: tcsetattr(0, TCSANOW, &oldtios);
			/* If a pending cut, mark it as extending */
			/* to the end of the stream */
			add_cut_end(1000000);
      		       record_program(outfname, stream_fd); Exit(0);

      case CMD_PLAYITALL:
        if (allow_playback == 0) {
	  fprintf(stderr, "\r\nYou can't playback with the -n flag\r\n");
	  break;
        }
        tcsetattr(0, TCSANOW, &oldtios); playback_program(stream_fd);
        tcsetattr(0, TCSANOW, &newtios); chunk= totalchunks - 1;
        paused= 1; break;

      case CMD_KILLCUT:
        if (!paused) {
	  fprintf(stderr, "\r\nCan only kill a cut in pause mode\r\n");
	  break;
        }
        switch (is_a_cut) {
          case IS_CUT_START: fprintf(stderr,
			"\r\nKilling cut starting at chunk %d\r\n", chunk);
	  kill_cut_start(chunk);
	  break;

        case IS_CUT_END: fprintf(stderr,
			"\r\nKilling cut ending at chunk %d\r\n", chunk);
	  kill_cut_end(chunk);
	  break;

        default:		/* If not at an endpoint, kill last cut */
	  cstart= kill_last_cut();
	  if (cstart == -1) fprintf(stderr, "\r\nNo cut to kill!\r\n");
	  else fprintf(stderr,
		  "\r\nKilling last cut which started at chunk %d\r\n", cstart);
          }
          break;

      case CMD_KILLALL: fprintf(stderr, "\r\nKilling all cuts\r\n");
      			kill_all_cuts(); break;

      case CMD_LISTCUTS: fprintf(stderr, "\r\nHere is the list of cuts:\r\n");
      		        list_cuts(); break;
    }

    /* For certain command during a pause, don't play another chunk */
    if (paused)
      switch (cmd) {
        case CMD_CSTART:
        case CMD_CEND:
        case CMD_PAUSE:
        case CMD_TOGSEECUT:
        case CMD_NONE: usleep(100000); goto here;	/* Yuk, a goto! */

	/* If not paused, go on to the next chunk */
    } else {
      chunk++;
#ifdef TIVO
      if (osd_dirty) { ClearOSD(); DrawOSD(); osd_dirty= 0; osd_chunk= 0; }
#endif
    }
  }
}

/* Print out usage information on the application & exit */
static void
usage(void)
{
  fprintf(stderr, "playitsam [options] inputfile [outputfile]\n");
  fprintf(stderr, "options:\n");
  fprintf(stderr, "\t -V print version number and exit\n");
  fprintf(stderr, "\t -v verbose\n");
  fprintf(stderr, "\t -n don't play the stream to a/v device\n");
  fprintf(stderr, "\t -p only play the stream, disallow editing\n");
#ifdef MPEG2DEC
  fprintf(stderr, "\t -c don't use CPU accelerations to play the video\n");
#endif
#ifdef TIVO
  fprintf(stderr, "\t -M restart myworld when playitsam exits\n");
  fprintf(stderr, "\t -D don't give on-screen user prompting\n");
#endif
  fprintf(stderr, "\t -h print this help information\n\n");
  fprintf(stderr, "input source is one of the following:\n");
  fprintf(stderr, "\t - means standard input (no skipping fwd/back)\n");
#ifdef TIVO
  fprintf(stderr, "\t NNNN means the fsid on MFS with that number\n");
#endif
  fprintf(stderr, "\t xxx.xxx.xxx.xxx:name means a file from a tyserver\n");
  fprintf(stderr, "\t anything else means a Unix filename\n\n");
  fprintf(stderr, "output source is one of the following:\n");
  fprintf(stderr, "\t - means standard output\n");
  fprintf(stderr, "\t xxx.xxx.xxx.xxx:name means a file to a tyserver\n");
  fprintf(stderr, "\t anything else means to a Unix filename\n\n");
  Exit(1);
}


int
main(int argc, char *argv[])
{
  char *fname;			/* Input file name */
  char *argflags;
  extern char *optarg;
  extern int optind;
  int c;

#ifdef TIVO
  argflags= "VvhnpMD";
#else
  argflags= "Vvhnpc";
#endif

  /* Get the various flags to the program */
  while ((c= getopt(argc, argv, argflags)) != -1) {
    switch (c) {
    case 'V': printf("\nPlayitsam $Revision: 1.69 $\n"); Exit(0);
    case 'v': verbose++; break;

    case 'n':
      if (allow_playback == 2) {
	fprintf(stderr, "Tell me, why did you try -n and -p?\n");
      }
      allow_playback= 0; break;
#ifdef MPEG2DEC
    case 'c': disable_accel= 1; break;
#endif
    case 'p':
      if (allow_playback == 0) {
	fprintf(stderr, "Tell me, why did you try -n and -p?\n");
      }
      allow_playback= 2; break;

    case 'M': exit_things |= EXIT_MYWORLD; break;

    case 'D': do_osd= 0; break;

    case 'h':
    default: usage();
    }
  }

  /* and skip over to the input argument */
  argc -= optind; argv += optind;
  if (argc < 1) { usage(); }
#ifdef TIVO

  /* First determine what sort of TiVo this is */
  determine_tivo_version();

  if (allow_playback) {
    /* We need to kill both EventSwitcher and MfsDaemon */
    c= pid_fromname("EventSwitcher");
    if (c > 0) {
      kill(c, SIGTERM);
      fprintf(stderr, "Killed EventSwitcher pid %d\n", c);
    }
    c= pid_fromname("MfsDaemon");
    if (c > 0) {
      kill(c, SIGTERM);
      fprintf(stderr, "Killed MfsDaemon pid %d\n", c);
    }
  }
#endif

  /*
   * Set the keyboard to raw mode & non-blocking, so we can get single
   * character commands
   */
  if (allow_playback != 2) {
    tcgetattr(0, &oldtios); memcpy(&newtios, &oldtios, sizeof(oldtios));
    set_nonblocking(0); cfmakeraw(&newtios); tcsetattr(0, TCSANOW, &newtios);
    exit_things |= EXIT_FIXKEYBOARD;
  }
  /* Try to open the input up */
  fname= argv[0];
  outfname= argv[1];
  stream_fd= open_input(fname);
  if (stream_fd == -1) {
    fprintf(stderr, "Can't open input file %s\n", fname);
    Exit(1);
  }
  /* Open the output devices if we are going to play things back */
  if (allow_playback) {
#ifdef TIVO
    opendevs();
    open_remote_device();
    set_high_process_priority();
#else
    openpipes(AUDIO_PROG, VIDEO_PROG);
#endif
  }
#ifdef TIVO
  if (do_osd) { SetupTextOSD(); ClearOSD(); DrawOSD(); }
#endif

  /* Discard the first chunk - it often contains garbage */
  if (inputtype != INPUT_FSID) { read_a_chunk(stream_fd, 0); }
  /* If they only want playback, give it to them */
  if (allow_playback == 2) { playback_program(stream_fd); Exit(0);
  }
  /* Otherwise, start editing the tystream */
  edit_program();

  Exit(0);
  return (0);
}
