#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <sys/stat.h>
#include <sched.h>
#include "playitsam.h"
#include "libmfs/mfs.h"
#include "osd.h"
#include "proto.h"
#include "tproto.h"
int snprintf(char *str, size_t size, const char *format, ...);

extern int exit_things;			/* Things to do at exit time */
extern int audio_fd;			/* fd for audio device */
extern int video_fd;			/* fd for video device */

static int remote_fd;                   /* fd for remote control device */
static int tivo_version;		/* What version of the TiVo is this? */

#define VERSION_UNKNOWN	 0		/* An unknown version */
#define VERSION_1_3	13		/* Version 1.3.x */
#define VERSION_2_5	25		/* Version 2.5.x */

void determine_tivo_version(void)
{
  FILE *zin;
  char line[1024];
  int i;

  struct kv {
	char *id_string;
	int tivo_type;
  } known_version[] = {
	{ "1.3.", VERSION_1_3 },
	{ "2.5.", VERSION_2_5 },
	{ NULL, VERSION_UNKNOWN }
  };


  tivo_version= VERSION_UNKNOWN;		/* Assume no result below */

  /* Open /etc/build-version */
  zin= fopen("/etc/build-version", "r");
  if (zin==NULL) {
    fprintf(stderr,
	"Can't open /etc/build-version, no idea what you're running\n");
    return;
  }
  
  /* Try and find an id on each input line */
  while (fgets(line, 1023, zin)) {
    for (i=0;known_version[i].tivo_type;i++) {
      if (strstr(line, known_version[i].id_string)) {
	tivo_version= known_version[i].tivo_type; break;
      }
    }
    if (tivo_version) break;
  }

  /* For now, a diagnostic for the user */
  switch (tivo_version) {
      case VERSION_1_3:
	fprintf(stderr, "You have a TiVo running version 1.3.x\n"); break;
      case VERSION_2_5:
	fprintf(stderr, "You have a TiVo running version 2.5.x\n"); break;
      default:
	fprintf(stderr, "I couldn't work out what sort of TiVo you have\n");
  }
  fclose(zin); return;
}




/* TODO: given a process name, find the matching pid.
 * Return the pid, or -1 if none.
 */
int pid_fromname(char *progname)
{
 DIR *d;
 FILE *cmdf;
 char *c, cmdline[80], cmdfile[1024];
 struct dirent *dp;

 d= opendir("/proc");				/* Open /proc directory */
 if (d==NULL) return(-1);

 while ((dp = readdir(d)) != NULL) {
   if (!isdigit(dp->d_name[0])) continue;	/* Skip non-digit filenames */

   snprintf(cmdfile, 1023, "/proc/%s/cmdline", dp->d_name);
   cmdf= fopen(cmdfile, "r");			/* Open /proc/NNN/cmdline */
   if (cmdf==NULL) continue;

   c=fgets(cmdline, 80, cmdf);			/* Read the command line */
   fclose(cmdf);				/* and close the file */
   if (c==NULL) continue;

   if (!strcmp(cmdline, progname)) {		/* If it matches what we have */
     (void)closedir(d);
     return(atoi(dp->d_name));			/* then return NNN as integer */
   }
 }
 (void)closedir(d);
 return(-1);					/* No match, return -1 */
}


/* Restart MyWorld. Thanks to D18C7DB for the tip and to the subscribers
 * of ExtractStream@yahoogroups.com who told me about version 2.5.
 */
void start_myworld(void)
{
  int fd, pid;
  char *momswitcher;
  
  /* First determine if we need to run momstart or switcherstart */
  switch (tivo_version) {
      case VERSION_1_3: momswitcher= "/tvbin/momstart"; break;
      case VERSION_2_5: momswitcher= "/tvbin/switcherstart"; break;
      default: return;		/* Not starting anything here! */
  }

  /* Now run momstart or switcherstart */
  pid=fork();
  switch (pid) {
    case -1: return;
    case 0: execl(momswitcher, momswitcher, "-m", NULL); exit(1);
    default: sleep(1);
  }

  /* Now run MyWorld */
  pid=fork();
  switch (pid) {
    case -1: return;
    case 0: fd=open("/dev/ttyS2", O_WRONLY);
	    if (fd<0) return;
	    close(1); dup2(fd, 1);
	    close(2); dup2(fd, 2);
	    close(fd);
	    execl("/tvbin/myworld", "/tvbin/myworld", NULL); exit(1);
    default: return;
  }
}


/*
 * Set the process on the TiVo to a high scheduling priority, see
 * http://www.linux.com/develop/man/2/sched_setscheduler/
 */
void set_high_process_priority(void)
{
  struct sched_param p;
  int err;

  p.sched_priority = 90;
  err = sched_setscheduler(0, SCHED_FIFO, &p);
}

/* Open the remote control device for reading */
void open_remote_device(void)
{
 remote_fd= open("/dev/ttyS0", O_RDONLY);
 set_nonblocking(remote_fd);
 exit_things |= EXIT_CLOSEREMDEV;
}

void close_remote(void)
{
 close(remote_fd);
}


/* On the TiVo, we can simply open up the video and audio device files */
void opendevs(void)
{
  video_fd = open("/dev/mpeg0v", O_WRONLY);
  audio_fd = open("/dev/mpeg0a", O_WRONLY);
  exit_things |= EXIT_CLOSEAVDEVS;
}


/*
 * The TiVo remote seems to send 6 bytes of keystoke down
 * data followed by 6 bytes of up data. The down data always
 * starts with the bytes 179 122 30, then 3 bytes describing
 * the button pressed. The keystroke up data is always the
 * 6 bytes: 244 0 0 0 0 0. Here is the list as far as I can tell:
 *
 * 179 122 30 127 127 0  244 0 0 0 0 0 	TiVo Central
 * 179 122 30 59 127 1   244 0 0 0 0 0 	Live TV
 * 179 122 30 47 127 1   244 0 0 0 0 0 	Up Arrow
 * 179 122 30 35 127 1   244 0 0 0 0 0 	<--
 * 179 122 30 39 127 1   244 0 0 0 0 0 	Down Arrow
 * 179 122 30 43 127 1   244 0 0 0 0 0	-->
 * 179 122 30 31 127 1   244 0 0 0 0 0 	Thumbs Down
 * 179 122 30 27 127 1   244 0 0 0 0 0 	Select
 * 179 122 30 23 127 1   244 0 0 0 0 0 	Thumbs Up
 * 179 122 30 127 126 2  244 0 0 0 0 0 	Record
 * 179 122 30 123 126 2  244 0 0 0 0 0	Play
 * 179 122 30 15 127 1   244 0 0 0 0 0 	Volume Up
 * 179 122 30 11 127 1   244 0 0 0 0 0 	Volume Down
 * 179 122 30 7 127 1    244 0 0 0 0 0 	Channel Up
 * 179 122 30 3 127 1    244 0 0 0 0 0	Channel Down
 * 179 122 30 119 126 2  244 0 0 0 0 0 	Back
 * 179 122 30 115 126 2  244 0 0 0 0 0 	Pause
 * 179 122 30 111 126 2  244 0 0 0 0 0 	Forward
 * 179 122 30 103 126 2  244 0 0 0 0 0	Go back 30 seconds
 * 179 122 30 107 126 2  244 0 0 0 0 0 	Slow
 * 179 122 30 99 126 2   244 0 0 0 0 0 	End
 * 179 122 30 95 126 2   244 0 0 0 0 0 	1
 * 179 122 30 91 126 2   244 0 0 0 0 0 	2
 * 179 122 30 87 126 2   244 0 0 0 0 0 	3
 * 179 122 30 83 126 2   244 0 0 0 0 0 	4
 * 179 122 30 79 126 2   244 0 0 0 0 0 	5
 * 179 122 30 75 126 2   244 0 0 0 0 0	6
 * 179 122 30 71 126 2   244 0 0 0 0 0 	7
 * 179 122 30 67 126 2   244 0 0 0 0 0 	8
 * 179 122 30 63 126 3   244 0 0 0 0 0 	9
 * 179 122 30 55 126 3   244 0 0 0 0 0 	Clear
 * 179 122 30 59 126 3   244 0 0 0 0 0 	0
 * 179 122 30 51 126 3   244 0 0 0 0 0 	Enter
 *
 *
 * Read a keystroke from the TiVo remote and return its
 * information. I don't convert to editing commands here
 * so that the buttons can have different meanings in
 * different modes.
 */
int read_a_remote_command(void)
{
  char ch;
  int ret, i, thismuch, cmd;

  while (1) {
    ret=read(remote_fd, &ch, 1);
    if (ret==-1) { return(REM_NONE); }
    if (ch!=30) continue;		/* Loop until just before key data */

    ret=read(remote_fd, &ch, 1);	/* Get 1st key byte */
    if (ret==-1) { return(REM_NONE); }

    thismuch=8;				/* Assume 8 bytes more after key id */
    switch(ch) {		       	/* Work out key id */
      case 3: cmd=  REM_CHAN_DOWN; break;
      case 7: cmd=  REM_CHAN_UP; break;
      case 11: cmd= REM_VOL_DOWN; break;
      case 15: cmd= REM_VOL_UP; break;
      case 23: cmd= REM_THUMBS_UP; break;
      case 27: cmd= REM_SELECT; break;
      case 31: cmd= REM_THUMBS_DOWN; break;
      case 35: cmd= REM_LEFT_ARROW; break;
      case 39: cmd= REM_DOWN_ARROW; break;
      case 43: cmd= REM_RIGHT_ARROW; break;
      case 47: cmd= REM_UP_ARROW; break;
      case 51: cmd= REM_ENTER; break;
      case 55: cmd= REM_CLEAR; break;
      case 63: cmd= REM_9; break;
      case 67: cmd= REM_8; break;
      case 71: cmd= REM_7; break;
      case 75: cmd= REM_6; break;
      case 79: cmd= REM_5; break;
      case 83: cmd= REM_4; break;
      case 87: cmd= REM_3; break;
      case 91: cmd= REM_2; break;
      case 95: cmd= REM_1; break;
      case 99: cmd= REM_END; break;
      case 103: cmd= REM_BACK30; break;
      case 107: cmd= REM_SLOW; break;
      case 111: cmd= REM_FORWARD; break;
      case 115: cmd= REM_PAUSE; break;
      case 119: cmd= REM_BACK; break;
      case 123: cmd= REM_PLAY; break;

      case 59:
	thismuch=7;
        ret=read(remote_fd, &ch, 1);
        if (ret==-1) { return(REM_NONE); }
	switch (ch) {
          case 126: cmd=REM_0; break;
          case 127: cmd=REM_LIVE_TV; break;
	  default:  cmd=REM_NONE;
	}
	break;

      case 127:
	thismuch=7;
        ret=read(remote_fd, &ch, 1);
        if (ret==-1) { return(REM_NONE); }
	switch (ch) {
          case 126: cmd=REM_RECORD; break;
          case 127: cmd=REM_TIVO_CENTRAL; break;
	  default:  cmd=REM_NONE;
	}
	break;

      default: cmd=REM_NONE;
    }

    /* Now read in the key up data to clean up */
    for (i=0;i<thismuch;i++) {
      ret=read(remote_fd, &ch, 1);
      if (ret==-1) break;
    }
    return(cmd);
  }					/* end of forever loop */
}
